# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines a `JobStore` that stores the `Job` information in memory.
This makes it easier for development or when persistent `Job`
data is not necessary."""

from typing import Optional
import datetime
import re
from . import JobStore, JobAlreadyExistsException, JobDoesNotExistException
from ..job import Job


class MemoryJobStore(JobStore):
    """Stores all `Job`s in memory, which will not persist.

    Attributes:
        _jobs (List[Job]): List of the `Job`s submitted to the store.
        _jobs_map (Dict[str, int]): Maps `job_id`s to an index in `_jobs`.
    """

    def __init__(self) -> None:
        super().__init__()
        self._jobs: list[Job] = []
        self._jobs_map: dict[str, int] = {}  # job_id -> index

    def teardown(self) -> None:
        super().teardown()
        self.remove_all_jobs()

    def add_job(self, job: Job, replace_existing: bool) -> None:
        if self.contains_job(job.id):
            if not replace_existing:
                raise JobAlreadyExistsException(job.id)

            self.update_job(job)
            return

        if job.next_run_time is None and job.active:
            job.next_run_time = job.trigger.get_next_fire_time(None, None)

        self._logger.debug(f'Add job ({job}) next run time: {job.next_run_time}')

        index = self._get_job_index(job.next_run_time)

        self._jobs.insert(index, job)
        self._jobs_map[job.id] = index

    def update_job(self, job: Job) -> None:
        old_index = self._jobs_map.get(job.id, None)
        if old_index is None:
            raise JobDoesNotExistException(job.id)

        self._logger.debug(f'Update job ({job}) next run time: {job.next_run_time}')

        old_job = self._jobs[old_index]

        if old_job.next_run_time == job.next_run_time:
            self._jobs[old_index] = job
        else:
            del self._jobs[old_index]
            del self._jobs_map[job.id]
            self.add_job(job, False)

    def remove_job(self, job_id: str) -> None:
        index = self._jobs_map.get(job_id, None)
        if index is None:
            raise JobDoesNotExistException(job_id)

        self._logger.debug(f'Delete job ({self._jobs[index]})')

        del self._jobs[index]
        del self._jobs_map[job_id]

    def remove_all_jobs(self) -> None:
        self._jobs = []
        self._jobs_map = {}

    def get_job(self, job_id: str) -> Job:
        index = self._jobs_map.get(job_id, None)
        if index is None:
            raise JobDoesNotExistException(job_id)
        return self._jobs[index]

    def get_jobs(self, pattern: str = None) -> list[Job]:
        if pattern is None:
            return self._jobs

        return [job for job in self._jobs if re.fullmatch(pattern, job.id)]

    def contains_job(self, job_id: str) -> bool:
        return job_id in self._jobs_map

    def _get_job_index(self, dt: Optional[datetime.datetime]) -> int:
        """Returns the index of a `Job` directly after the provided datetime.

        Args:
            dt (Optional[datetime.datetime]): Used to find the next `Job` directly after this
                datetime.

        Returns:
            int: The index of the first `Job` after the provided datetime, or the next index to be
            filled in the list.
        """
        index = -1

        if dt is not None:
            for i, job in enumerate(self._jobs):
                if job.next_run_time is not None and dt < job.next_run_time:
                    index = i
                    break

        if index == -1:
            index = len(self._jobs)

        return index

    def _get_stored_instance_id(self) -> int:
        return 0

    def _save_instance_id(self, instance_id: int) -> None:
        pass



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines a `JobStore` that utilizes an SQLAlchemy database.
This allows for persistent data storage.

Two tables are utilized:
* `yapsched_jobs`: Stores `Job` data.
* `yapsched_last_instance_id: Stores the last instance id used.

It utilizes Python's `pickle` module to serialize `Job` data before inserting into the database."""

from decimal import Decimal
from typing import Any, Callable, Optional, TypeVar, cast
from datetime import datetime
import pickle
import re
from sqlalchemy import create_engine, Column, Unicode, Float, LargeBinary, Integer
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session, Session
from . import JobStore, JobDoesNotExistException, JobAlreadyExistsException
from ..job import Job


Base = declarative_base()
SessionCls = scoped_session(sessionmaker())

F = TypeVar('F', bound=Callable[..., Any])  # pylint: disable=invalid-name

def db_user(func: F) -> F:
    """Wrapper function to either utilize an SQLAlchemy session or create a new one.

    Args:
        func (F): Wrapped to function to call after setting up the session.
    """
    def wrapper(self: 'SqlAlchemyJobStore', *args: Any, **kwargs: Any) -> Any:
        session: Optional[Session] = kwargs.get('session', None)
        session_exists = True

        if session is None:
            session = SessionCls()
            kwargs['session'] = session
            session_exists = False

        result = func(self, *args, **kwargs)

        if not session_exists:
            session.commit()
            session.close()

        return result
    return cast(F, wrapper)


class DbJob(Base):
    """Defines an SQLAlchemy table for `Job`s and maps onto it."""
    __tablename__ = 'yapsched_jobs'

    id = Column(Unicode(191, _warn_on_bytestring=False), primary_key=True)
    next_run_time = Column(Float(25), index=True)
    job_state = Column(LargeBinary, nullable=False)


class LastInstanceId(Base):
    """Defines an SQLAlchemy table for the `Scheduler`'s last instance id and maps onto it."""
    __tablename__ = 'yapsched_last_instance_id'

    row = Column(Integer, primary_key=True)
    id = Column(Integer)


class SqlAlchemyJobStore(JobStore):
    """A `JobStore` that utilizes an SQLAlchemy database.

    Attributes:
        pickle_protocol (int): Determines what protocol version to use when pickling.
        engine (Engine): The SQLAlchemy engine to utilize for the database.
    """

    def __init__(self, url: str = None, engine: Engine = None, engine_options: Optional[dict] = None,
                 pickle_protocol: int = pickle.HIGHEST_PROTOCOL) -> None:
        """Initializes a `SQLAlchemyJobStore`.

        Either `url` or `engine` must be defined to establish an `Engine` for SQL Alchemy to use.

        Args:
            url (str, optional): The database URL used to create an `Engine`. Defaults to None.
            engine (Engine, optional): An `Engine` used by SQLAlchemy. Defaults to None.
            engine_options (Optional[Dict]): Options used to create an `Engine` if
                `engine` is None. Defaults to None.
            pickle_protocol (int, optional): Determines what protocol version to use when pickling.
                Defaults to pickle.HIGHEST_PROTOCOL.

        Raises:
            ValueError: If both the `url` and `engine` are undefined.
        """
        super().__init__()
        self.pickle_protocol = pickle_protocol

        if engine is not None:
            self.engine = engine
        elif url is not None:
            self.engine = create_engine(url, **(engine_options or {}))
        else:
            raise ValueError('Need either "engine" or "url" defined')

        SessionCls.configure(bind=self.engine)

    def setup(self) -> None:
        super().setup()
        Base.metadata.create_all(self.engine)

    def teardown(self) -> None:
        super().teardown()
        self.engine.dispose()

    @db_user
    def add_job(self, job: Job, replace_existing: bool = True, session: Optional[Session] = None) -> None:
        """Adds a `Job` to store and track.

        Args:
            job (Job): A 'Job' to store and track.
            replace_existing (bool, optional): Determines if a `Job` matching the one provided
                should be replaced. Defaults to True.
            session (Session, optional): The SQLAlchemy Session to use. Defaults to None.

        Raises:
            JobAlreadyExistsException: If a `Job` matching the one provided is found, but
                `replace_existing` is False.
        """
        if self.contains_job(job.id):
            if not replace_existing:
                raise JobAlreadyExistsException(job.id)

            self.update_job(job, session=session)
            return

        if job.next_run_time is None and job.active:
            job.next_run_time = job.trigger.get_next_fire_time(None, None)

        self._logger.debug(f'Add job ({job}) next run time: {job.next_run_time}')

        next_run_time = cast(Optional[Decimal], job.get_next_run_time_ts())  # See https://github.com/dropbox/sqlalchemy-stubs/issues/178
        new_job = DbJob(id=job.id,
                        next_run_time=next_run_time,
                        job_state=pickle.dumps(job.__getstate__(), self.pickle_protocol))

        session.add(new_job)  # type: ignore

    @db_user
    def update_job(self, job: Job, session: Optional[Session] = None) -> None:
        """Updates an existing `Job` matching the `job.id` with new data.

        Args:
            job (Job): Find a `Job` matching this `job.id` and replace it with this.
            session (Session, optional): An SQLAchemy Session. Defaults to None.
        """
        existing_job = self._get_db_job(job.id, session)  # type: ignore

        if job.next_run_time is None and job.active:
            if existing_job.next_run_time is None:
                job.next_run_time = job.trigger.get_next_fire_time(None, None)
            else:
                tz = self._scheduler.tz  # type: ignore
                job.next_run_time = datetime.fromtimestamp(float(existing_job.next_run_time), tz=tz)

        self._logger.debug(f'Update job ({job}) next run time: {job.next_run_time}')

        next_run_time = cast(Optional[Decimal], job.get_next_run_time_ts())  # See https://github.com/dropbox/sqlalchemy-stubs/issues/178
        existing_job.next_run_time = next_run_time
        existing_job.job_state = pickle.dumps(job.__getstate__(), self.pickle_protocol)

    @db_user
    def remove_job(self, job_id: str, session: Optional[Session] = None) -> None:
        """Removes a single `Job`.

        Args:
            job_id (str): Used to find the `Job` to remove.
            session (Session, optional): An SQLAlchemy Session. Defaults to None.
        """
        existing_job = self._get_db_job(job_id, session)  # type: ignore
        self._logger.debug(f'Delete job ({self._reconstitute_job(existing_job)})')
        session.delete(existing_job)  # type: ignore

    @db_user
    def remove_all_jobs(self, session: Optional[Session] = None) -> None:
        """Removes all `Job`s.

        Args:
            session (Session, optional): An SQLAlchemy Session. Defaults to None.
        """
        session.query(DbJob).delete()  # type: ignore

    @db_user
    def get_job(self, job_id: str, session: Optional[Session] = None) -> Job:
        """Returns a `Job` matching the `job_id`.

        Args:
            job_id (str): Used to find the `Job` to return.
            session (Session, optional): An SQLAlchemy Session. Defaults to None.

        Returns:
            Job: The `Job` associated with the `job_id`.
        """
        existing_job = self._get_db_job(job_id, session)  # type: ignore
        return self._reconstitute_job(existing_job)

    @db_user
    def get_jobs(self, pattern: str = None, session: Optional[Session] = None) -> list[Job]:
        """Returns a list of `Job`s matching the pattern, or all if no pattern.

        Args:
            pattern (str, optional): A regular expression pattern to match against `Job` ids.
                Defaults to None.
            session (Session, optional): An SQLAlchemy Session. Defaults to None.

        Returns:
            List[Job]: A list of `Job`s matching the pattern, or all if no pattern.
        """
        db_jobs = session.query(DbJob).order_by(DbJob.next_run_time.asc()).all()  # type: ignore

        if pattern is not None:
            compiled_pattern = re.compile(pattern)
            db_jobs = list(filter(lambda db_job: compiled_pattern.fullmatch(db_job.id), db_jobs))

        return list(map(lambda db_job: self._reconstitute_job(db_job), db_jobs))

    @db_user
    def contains_job(self, job_id: str, session: Optional[Session] = None) -> bool:
        """Returns if the `JobStore` currently has a `Job`.

        Args:
            job_id (str): Used to find a `Job`.
            session (Session, optional): An SQLAlchemy Session. Defaults to None.

        Returns:
            bool: True of the `JobStore` currently has a `Job` with the `job_id`, False otherwise.
        """
        try:
            self._get_db_job(job_id, session)  # type: ignore
            return True
        except JobDoesNotExistException:
            return False

    @staticmethod
    def _get_db_job(job_id: str, session: Session) -> DbJob:
        """Returns a `Job` from the database.

        Args:
            job_id (str): Used to find a `Job`.
            session (Session): An SQLAlchemy Session.

        Raises:
            JobDoesNotExistException: If the `job_id` does not match any stored `Job`.

        Returns:
            DbJob: A stored, pickled version of the `Job`.
        """
        existing_job = session.query(DbJob).get(job_id)  # type: ignore
        if existing_job is None:
            raise JobDoesNotExistException(job_id)
        return existing_job

    @staticmethod
    def _reconstitute_job(db_job: DbJob) -> Job:
        """Rebuilds a `Job` from a stored `DbJob` entry.

        Args:
            db_job (DbJob): The stored, pickled version of a `Job`.

        Returns:
            Job: A `Job` rebuilt using the data from `DbJob`.
        """
        job_state = pickle.loads(db_job.job_state)
        job = object.__new__(Job)
        job.__setstate__(job_state)
        return job

    @db_user
    def _get_stored_instance_id(self, session: Optional[Session] = None) -> int:
        """Returns the last stored instance id.

        Args:
            session (Session, optional): An SQLAlchemy Session. Defaults to None.

        Returns:
            int: The last stored instance id from the `LastInstanceId` table.
        """
        last_instance_id = session.query(LastInstanceId).get(0)  # type: ignore
        if last_instance_id is None:  # type: ignore
            return 0
        return last_instance_id.id

    @db_user
    def _save_instance_id(self, instance_id: int, session: Optional[Session] = None) -> None:
        """Saves an instance id.

        Args:
            instance_id (int): The instance id to save to the database.
            session (Session, optional): An SQLAlchemy Session. Defaults to None.
        """
        last_instance_id = session.query(LastInstanceId).get(0)  # type: ignore
        if last_instance_id is None:
            last_instance_id = LastInstanceId(row=0, id=instance_id)
            session.add(last_instance_id)  # type: ignore
        else:
            last_instance_id.id = instance_id

        self._logger.debug(f'saved new instance ID ({instance_id})')



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an Abstract Base Class for `JobStore`."""

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, TYPE_CHECKING
from ..util import get_class_logger
from ..job import Job

if TYPE_CHECKING:
    from ..scheduler import Scheduler

MIN_INSTANCE_ID = 10000
MAX_INSTANCE_ID = 99999


class JobStore(ABC):
    """Abstract Base Class for `JobStore`s.
    
    Utilizes instance ids to track a `Job` and its associated data when running.

    Attributes:
        _logger (logging.Logger): The class' logger.
        _scheduler (Scheduler): The parent `Scheduler`.
        last_instance_id (int): The last instance id assigned.
    """

    def __init__(self) -> None:
        self._logger = get_class_logger(self)
        self._scheduler: Optional['Scheduler'] = None

        self.last_instance_id = 0

    def setup(self) -> None:
        """Any basic setup necessary for the `JobStore`."""
        pass

    def teardown(self) -> None:
        """Cleans up `Job`s and their instances."""
        pass

    def get_due_jobs(self, latest: datetime) -> list[Job]:
        """Returns a list of all `Job`s that are due to run.

        Args:
            latest (datetime): Filters out all `Job`s scheduled to run after this time.

        Returns:
            List[Job]: A list of `Job`s scheduled to run before `latest`.
        """
        jobs = self.get_jobs()
        pending = list(filter(lambda job: job.active and job.next_run_time is not None and\
            job.next_run_time <= latest, jobs))
        return pending

    def get_next_run_time(self) -> Optional[datetime]:
        """Returns the next upcoming run time of the soonest `Job`, if any is scheduled."""
        jobs = self.get_jobs()
        return jobs[0].next_run_time if jobs else None

    @abstractmethod
    def add_job(self, job: Job, replace_existing: bool) -> None:
        """Adds a `Job` to store and track.

        Args:
            job (Job): A `Job` to store and track.
            replace_existing (bool): Determines if a `Job` matching the one provided should be
                replaced.
        """
        pass

    @abstractmethod
    def update_job(self, job: Job) -> None:
        """Updates an existing `Job` matching `job.id` with the new data.

        Args:
            job (Job): Used to update the current `Job` matching the provided `Job`'s id.
        """
        pass

    @abstractmethod
    def remove_job(self, job_id: str) -> None:
        """Removes a single `Job`.

        Args:
            job_id (str): Identifier for the `Job` to remove.
        """
        pass

    @abstractmethod
    def remove_all_jobs(self) -> None:
        """Removes all `Job`s."""
        pass

    @abstractmethod
    def get_job(self, job_id: str) -> Job:
        """Returns the `Job` associated with the `job_id`.

        Args:
            job_id (str): Identifier for the `Job` to retrieve.

        Returns:
            Job: The `Job` associated with the `job_id`.
        """
        pass

    @abstractmethod
    def get_jobs(self, pattern: str = None) -> list[Job]:
        """Returns a list of `Jobs` matching the pattern, or all if no pattern.

        Args:
            pattern (str): A regular expression pattern to match against `Job` ids.

        Returns:
            List[Job]: A list of `Job`s matching the pattern, or all if no pattern.
        """
        pass

    @abstractmethod
    def contains_job(self, job_id: str) -> bool:
        """Returns if the `JobStore` currently has a `Job`.

        Args:
            job_id (str): The identifier to look for.

        Returns:
            bool: True of the `JobStore` currently has a `Job` with the `job_id`, False otherwise.
        """
        pass

    def get_new_instance_id(self) -> int:
        """Returns the next instance id to use.

        Instance ids are in the range `[MIN_INSTANCE_ID, MAX_INSTANCE_ID]` and loop back around
        when the max is reached.

        Returns:
            int: The next instance id to use.
        """
        if self.last_instance_id == 0:
            self._logger.debug('first time generating a new instance ID. getting last one from jobstore')
            self.last_instance_id = self._get_stored_instance_id()
            if self.last_instance_id == 0:
                self._logger.debug(f'no last instance ID in jobstore. using minimum value ({MIN_INSTANCE_ID})')
                # we use MIN - 1 here because of the increment a few lines down
                self.last_instance_id = MIN_INSTANCE_ID - 1
            else:
                self._logger.debug(f'got last instance ID from jobstore ({self.last_instance_id})')

        if self.last_instance_id < MIN_INSTANCE_ID - 1 or self.last_instance_id > MAX_INSTANCE_ID:
            self._logger.error(f'Last instance ID ({self.last_instance_id}) was out of acceptable range '
                               f'([{MIN_INSTANCE_ID}, {MAX_INSTANCE_ID}])! Resetting to minimum value '
                               f'({MIN_INSTANCE_ID})')

            # we use MIN - 1 here because of the increment in the next line of code
            self.last_instance_id = MIN_INSTANCE_ID - 1

        self.last_instance_id += 1

        if self.last_instance_id > MAX_INSTANCE_ID:
            self._logger.debug(f'new instance ID at max ({MAX_INSTANCE_ID}), resetting to min ({MIN_INSTANCE_ID})')
            self.last_instance_id = MIN_INSTANCE_ID

        self._save_instance_id(self.last_instance_id)

        self._logger.debug(f'generated a new instance ID ({self.last_instance_id})')

        return self.last_instance_id

    @abstractmethod
    def _get_stored_instance_id(self) -> int:
        """Returns the last used instance id that was stored.
        
        Returns:
            int: The last used instance id that was stored.
        """
        pass

    @abstractmethod
    def _save_instance_id(self, instance_id: int) -> None:
        """Saves the last used instance id."""
        pass


class JobAlreadyExistsException(Exception):
    """Raised when a `Job` with the provided identifier already exists."""

    def __init__(self, job_id: str) -> None:
        """Initializes a JobAlreadyExistsException.

        Args:
            job_id (str): Identifier for a `Job` that was already in use.
        """
        super().__init__(f'Job "{job_id}" already exists')


class JobDoesNotExistException(Exception):
    """Raised when a `Job` identifier was used to retrieve a `Job`, but it did not exist."""

    def __init__(self, job_id: str) -> None:
        """Initializes a JobDoesNotExistException.

        Args:
            job_id (str): Identifier for a `Job` which was not in the `JobStore`.
        """
        super().__init__(f'Job "{job_id}" does not exist')



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines and describes the types of events that may occur in the `Scheduler`.

This includes events from both `Job`s and the `Scheduler` itself.
`Event`s are integers but can be seen as bitstring flags, such that bitwise operations
can be performed on them.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional
import traceback

EVENT_SCHEDULER_STARTED = 2 ** 0
EVENT_SCHEDULER_STOPPED = 2 ** 1
EVENT_SCHEDULER_PAUSED = 2 ** 2
EVENT_SCHEDULER_RESUMED = 2 ** 3
EVENT_ALL_JOBS_REMOVED = 2 ** 4
EVENT_JOB_ADDED = 2 ** 5
EVENT_JOB_REMOVED = 2 ** 6
EVENT_JOB_MODIFIED = 2 ** 7
EVENT_JOB_EXECUTED = 2 ** 8
EVENT_JOB_ERROR = 2 ** 9
EVENT_JOB_MISSED = 2 ** 10
EVENT_JOB_SUBMITTED = 2 ** 11
EVENT_JOB_MAX_INSTANCES = 2 ** 12
EVENT_JOB_TERMINATED = 2 ** 13

# UPDATE THIS WHENEVER AN EVENT IS ADDED/REMOVED!
# if the last event is 2 ** x, this should be 2 ** (x + 1) - 1
EVENT_ALL = 2 ** 14 - 1


@dataclass
class Event:
    """Base class to describe events.

    Attributes:
        code (int): Represents a bitstring flag.
    """
    code: int


class SchedulerEvent(Event):
    """Describes a `Scheduler` event."""
    pass


@dataclass
class JobEvent(Event):
    """Base class to describe any `Job` related event.

    Attributes:
        job_id (str): Unique identifier for this `Job`.
    """
    job_id: str


class JobExecutionEvent(JobEvent):
    """Describes the status and properties of a `Job` that was executed.

    Attributes:
        instance_id (int): The `Job`'s process instance identifier.
        run_time (datetime): The time when the `Job` was executed.
        success (bool): If the `Job` ran successfully.
        retval (Any): The return value of the `Job`'s `func`.
        exc (Exception): An exception to raise if the `Job` did not run successfully.
        formatted_exc (str): A string format version of the exception and its trace.
    """

    def __init__(self, code: int, job_id: str, instance_id: int, run_time: datetime, success: bool, retval: Any = None,
                 exc: Optional[Exception] = None) -> None:
        """Initializes a `JobExecutionEvent`.

        Args:
            code (int): A bitstring flag for this event.
            job_id (str): The `Job`'s unique identifier.
            instance_id (int): The `Job`'s process instance identifier.
            run_time (datetime): When the `Job` was executed.
            success (bool): If the `Job` executed succesfully.
            retval (Any, optional): The return value of the `Job`'s `func`. Defaults to None.
            exc (Exception, optional): An exception that was raised by the `Job`. Defaults to None.
        """
        super().__init__(code, job_id)

        self.instance_id = instance_id
        self.run_time = run_time
        self.success = success
        self.retval = retval
        self.exc = exc

        self.formatted_exc: Optional[str]
        if exc is not None:
            self.formatted_exc = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__)).rstrip()
        else:
            self.formatted_exc = None



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The main interface for this package, the `Scheduler` manages a set of `Job`s.

A `Scheduler` maintains a `JobStore` of `Job`s to run, which are run through an `Executor`. The
`Job`s are run in a multiprocessing fashion.
Listeners allow for the additional execution of methods depending on `Scheduler` or `Job`
`Event`s.

The `Scheduler` is run on a thread with its `_main_loop` method.
"""

from enum import Enum, auto
from typing import Any, ClassVar, Optional, Callable
import threading
import uuid
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo
import logging
import zoneinfo
from .job import Job, JobFunction, JobCallback, JobErrorCallback
from . import executor, triggers, events
from .jobstores import JobDoesNotExistException, JobStore, memory

logger = logging.getLogger(__name__)


class State(Enum):
    """Enum describing the state of the `Scheduler`."""
    STOPPED = auto()
    RUNNING = auto()
    PAUSED = auto()


class Scheduler:
    """Manages the execution of provided Jobs via their respective scheduled triggers.

    The `Scheduler` has three states:
    * RUNNING
    * PAUSED
    * STOPPED

    Each `Job` can be added or removed from the provided `JobStore`, or modified in place.
    Listeners allow for certain actions to be taken when `Scheduler` or `Job` events occur.

    Attributes:
        _executor (executor.Executor): Handles the execution of Jobs.
        _executor_lock (threading.RLock): Reentrant lock to control access to the `_executor`.
        _jobstore (Optional[jobstores.JobStore]): JobStore for the Scheduler's jobs.
            Defaults to None.
        _jobstore_lock (threading.RLock): Reentrant lock to control access to the `_jobstore`.
        _listeners (List[Tuple(Callable, int)]): List of functions to be called upon a certain
            event occuring.
        _listeners_lock (threading.RLock): Reentrant lock to control access to the `_listeners`.
        state (State): Describes the Scheduler's current state.
        tz (zoneinfo): The timezone for the Scheduler to operate in. Defaults to
            timezone.utc.
    """
    _event: ClassVar[threading.Event]
    _thread: ClassVar[threading.Thread]

    @classmethod
    def set_thread(cls, thread: threading.Thread) -> None:
        """Sets the class `_thread` variable."""
        cls._thread = thread

    @classmethod
    def set_event(cls) -> None:
        """Sets the class `_event` variable."""
        cls._event = threading.Event()

    def __init__(self, pool_size: int = 10, jobstore: Optional[JobStore] = None, tz: zoneinfo = ZoneInfo('UTC')) -> None:
        """Initializes the `Scheduler`.

        Args:
            pool_size (int): Determines the pool size for the job `Executor`. Defaults
                to 10.
            jobstore (Optional[jobstores.JobStore]): A `JobStore` to manage the `Scheduler`'s
                `Job`s. Defaults to None.
            tz (zoneinfo): The timezone for the `Scheduler` to operate in. Defaults to
                timezone.utc.

        Raises:
            ValueError: If the provided timezone is None.
        """
        if tz is None:
            raise ValueError('tz must not be None')

        self._executor = executor.Executor(pool_size, self)
        self._executor_lock = threading.RLock()

        self._jobstore_lock = threading.RLock()

        self._listeners: list[tuple[Callable[[events.Event], None], int]] = []
        self._listeners_lock = threading.RLock()

        self.state = State.STOPPED

        self.tz: zoneinfo = tz

        self.configure(jobstore=jobstore, tz=tz)

    def configure(self, jobstore: Optional[JobStore] = None, tz: zoneinfo = None) -> None:
        """Configures the `_jobstore` and `tz` after initialization.

        Args:
            jobstore (Optional[jobstores.JobStore]): `JobStore` for the `Scheduler`'s `Job`s.
                Defaults to None.
            tz (zoneinfo): The timezone for the `Scheduler` to operate in.
        """
        if jobstore is None:
            jobstore = memory.MemoryJobStore()

        jobstore._scheduler = self
        self._jobstore = jobstore

        if tz is not None:
            self.tz = tz

    def start(self, paused: bool = False) -> None:
        """Starts the `Scheduler`.

        Args:
            paused (bool): Determines if the `Scheduler` should start paused or running.
                Defaults to False.

        Raises:
            SchedulerAlreadyRunningException: If the `Scheduler` is already running.
        """
        if self.state != State.STOPPED:
            raise SchedulerAlreadyRunningException

        self.set_event()

        with self._executor_lock:
            self._executor.setup()

        with self._jobstore_lock:
            self._jobstore.setup()

        self.add_listener(self._internal_listener)

        self.state = State.PAUSED if paused else State.RUNNING
        logger.info('Scheduler started')
        self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_STARTED))

        if not paused:
            self.wakeup()

        thread = threading.Thread(target=self._main_loop, name='yapsched', daemon=True, args=(self._event,))
        self.set_thread(thread)
        self._thread.start()

    def shutdown(self, wait: bool = True) -> None:
        """Shuts down the `Scheduler`.

        Args:
            wait (bool): Determines if it should wait for `Job` processes to stop.
                Defaults to True.

        Raises:
            SchedulerNotRunningException: If the `Scheduler` is not running.
        """
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException

        logger.info('Shutting down...')

        self.state = State.STOPPED

        logger.debug('Closing thread')
        self._event.set()
        self._thread.join()

        logger.debug('Deleting thread')
        delattr(Scheduler, '_thread')

        logger.debug('Tearing down executor')
        with self._executor_lock:
            self._executor.teardown(wait)

        logger.debug('Tearing down jobstore')
        with self._jobstore_lock:
            self._jobstore.teardown()

        self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_STOPPED))

        self.remove_listener(self._internal_listener)

        logger.info('Scheduler shutdown')

    def pause(self) -> None:
        """Pauses the `Scheduler`.

        Raises:
            SchedulerNotRunningException: If the `Scheduler` is not running.
        """
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException
        elif self.state == State.RUNNING:
            self.state = State.PAUSED
            logger.info('Scheduler paused')
            self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_PAUSED))

    def resume(self) -> None:
        """Resumes the `Scheduler`.

        Raises:
            SchedulerNotRunningException: If the `Scheduler` is stopped.
        """
        if self.state == State.STOPPED:
            raise SchedulerNotRunningException
        elif self.state == State.PAUSED:
            self.state = State.RUNNING
            logger.info('Scheduler resumed')
            self._dispatch_event(events.SchedulerEvent(events.EVENT_SCHEDULER_RESUMED))
            self.wakeup()

    @property
    def running(self) -> bool:
        """bool: Determines if the `Scheduler` is currently running (RUNNING or PAUSED)."""
        return self.state != State.STOPPED

    def wakeup(self) -> None:
        """Wakes the `Scheduler` up to process events."""
        self._event.set()

    def add_job(self,
                func: JobFunction,
                args: Optional[tuple] = None,
                kwargs: Optional[dict[str, Any]] = None,
                dynamic_args: Optional[dict[str, Callable]] = None,
                identifier: Optional[str] = None,
                description: Optional[str] = None,
                coalesce: bool = False,
                max_instances: int = -1,
                replace_existing: bool = False,
                trigger: Optional[triggers.Trigger] = None,
                next_run_time: Optional[datetime] = None,
                active: bool = True,
                callback: Optional[JobCallback] = None,
                error_callback: Optional[JobErrorCallback] = None) -> Job:
        """Adds a `Job` to the `Jobstore`.

        Args:
            func (JobFunction): The function to be called when the job executes.
            args (Optional[List]): The arguments to pass to the `func`. Defaults to None.
            kwargs (Optional[Dict]): The keyword arguments to pass to the `func`. Defaults to None.
            dynamic_args (Optional[Dict[str, Callable]]): A dictionary of argument names mapping to
                functions that will be executed upon job execution. The return values of the
                functions will be passed to func as keyword arguments. Defaults to None.
            identifier (Optional[str]): A unique ID for the job. Defaults to None.
            description (Optional[str]): The description of the job. Defaults to None.
            coalesce (bool): If true, it will only run once when several run timnes are due.
                Defaults to False.
            max_instances (int): The maximum number of concurrently running instances of the job.
                -1 means no limit, 0 means disabled. Defaults to -1.
            replace_existing (bool): Determines if a currently existing job of this id should be
                replaced.  Defaults to False.
            trigger (Optional[triggers.Trigger]): The trigger used to determine when the job will
                execute. Defaults to None.
            next_run_time (Optional[datetime]): When the job should run next. Defaults to None.
            active (bool): If the job should be marked as active. Defaults to True.
            callback (Optional[JobCallback]): A function to call if `func` executes
                successfully. It must accept three arguments: `job_id`, `instance_id`, and `retval`.
                The `retval` argument will be the return value of `func`. Defaults to None.
            error_callback (Optional[JobErrorCallback]): A function to call if `func` raises an
                exception. It must accept three arguments: `job_id`, `instance_id`, and `exc`.
                The `exc` argument will be the exception instance. Defaults to None.

        Raises:
            ValueError: If there is no next run time and the `Job` isn't active.

        Returns:
            Job: An instance of the `Job` created by the provided arguments.
        """
        if next_run_time is not None and not active:
            raise ValueError('Job can\'t be inactive and have a non-null next_run_time!')

        if identifier is None:
            identifier = str(uuid.uuid4())

        if trigger is None:
            trigger = triggers.DateTrigger(datetime.now(tz=self.tz))

        trigger.tz = self.tz

        job = Job(func,
                  identifier,
                  tuple(args) if args is not None else (),
                  kwargs if kwargs is not None else {},
                  dynamic_args if dynamic_args is not None else {},
                  trigger,
                  next_run_time=next_run_time,
                  description=description,
                  coalesce=coalesce,
                  max_instances=max_instances,
                  active=active,
                  callback=callback,
                  error_callback=error_callback)

        with self._jobstore_lock:
            self._jobstore.add_job(job, replace_existing)

        logger.info(f'Added job {identifier}')
        self._dispatch_event(events.JobEvent(events.EVENT_JOB_ADDED, identifier))

        if self.state == State.RUNNING and active:
            self.wakeup()

        return job

    def modify_job(self, job_id: str, **changes: Any) -> Job:
        """Modifies the `Job` with the given `job_id`.

        Changes is a dictionary containing mappings of `Job` attributes
        to new values.

        Args:
            job_id (str): Identifier for the `Job` to modify.
            **changes (dict[str, Any]): Keyword arguments to modify the job with.
        Raises:
            ValueError: If the newly provided next run time is None and the `Job` is not active.

        Returns:
            Job: The `Job` instance after being updated.
        """
        with self._jobstore_lock:
            job = self.get_job(job_id)

            new_next_run_time: Optional[datetime] = changes.get('next_run_time', None)

            if new_next_run_time and not isinstance(new_next_run_time, datetime):
                raise ValueError(f'Expected changes["next_run_time"] to be a datetime, but got '
                                 f'{type(new_next_run_time)}')
            elif new_next_run_time is None:
                new_next_run_time = job.next_run_time

            new_active: Optional[bool] = changes.get('active', None)

            if new_active and not isinstance(new_active, bool):
                raise ValueError(f'Expected changes["active"] to be a bool, but got '
                                 f'{type(new_active)}')
            elif new_active is None:
                new_active = job.active

            if new_next_run_time is not None and not new_active:
                raise ValueError('Job can\'t be inactive and have a non-null next_run_time!')

            job._modify(**changes)

            if job.trigger is None:
                job.trigger = triggers.DateTrigger(datetime.now(tz=self.tz))

            job.trigger.tz = self.tz

            self._jobstore.update_job(job)

        self._dispatch_event(events.JobEvent(events.EVENT_JOB_MODIFIED, job_id))

        if self.state == State.RUNNING:
            self.wakeup()

        return job

    def get_job(self, job_id: str) -> Job:
        """Returns the `Job` with the given `job_id`.

        Args:
            job_id (str): The identifier for the `Job` to get.

        Returns:
            Job: The `Job` associated with the given `job_id`.
        """
        with self._jobstore_lock:
            return self._jobstore.get_job(job_id)

    def get_jobs(self, pattern: str = None) -> list[Job]:
        """Returns all `Job`s.

        Args:
            pattern (str, optional): A regular expression pattern to match `Job` ids against.
                Defaults to None.

        Returns:
            List[Job]: A list of all the `Job`s.
        """
        with self._jobstore_lock:
            return self._jobstore.get_jobs(pattern)

    def remove_job(self, job_id: str) -> None:
        """Removes a single `Job` instance.

        Args:
            job_id (str): The identifier for the `Job` to remove.
        """
        with self._jobstore_lock:
            self._jobstore.remove_job(job_id)

        logger.info(f'Removed job {job_id}')
        self._dispatch_event(events.JobEvent(events.EVENT_JOB_REMOVED, job_id))

    def terminate_job(self, instance_id: int) -> None:
        """Terminates a single `Job` instance.

        Args:
            instance_id (int): The instance id for the `Job`.
        """
        with self._executor_lock:
            job_id = self._executor.terminate_job(instance_id)
            logger.info(f'Terminated job {job_id}')
            self._dispatch_event(events.JobEvent(events.EVENT_JOB_TERMINATED, job_id))

    def trigger_job(self, job_id: str) -> int:
        """Triggers a `Job` to run immediately.

        Args:
            job_id (str): The identifier for the `Job` to trigger.

        Returns:
            int: The instance id for the `Job`'s execution.
        """
        job = self.get_job(job_id)
        run_time = datetime.now(tz=self.tz)

        try:
            instance_id = self._executor.submit_job(job, run_time)
        except executor.MaxJobInstancesReachedException:
            self._dispatch_event(events.JobEvent(events.EVENT_JOB_MAX_INSTANCES, job_id))
            raise

        return instance_id

    def get_job_instance_ids(self, job_id: str) -> list[int]:
        """Returns the instance ids for the given `Job`.

        Args:
            job_id (str): The identifier for the `Job` with running instances.

        Raises:
            jobstores.JobDoesNotExistException: If the `Job` does not currently exist.

        Returns:
            List[int]: A list of instance ids associated with the given `job_id`.
        """
        with self._jobstore_lock:
            if not self._jobstore.contains_job(job_id):
                raise JobDoesNotExistException(job_id)

        with self._executor_lock:
            return self._executor.get_job_instance_ids(job_id)

    def get_job_with_instance_id(self, instance_id: int) -> Job:
        """Returns the `Job` associated with the given instance id.

        Args:
            instance_id (int): The id for the instance we're interested in.

        Returns:
            Job: The `Job` associated with the given instance id.
        """
        with self._executor_lock:
            return self._executor.get_job(instance_id)

    def get_running_jobs(self) -> list[executor.Instance]:
        """Returns a list of all `Job`s that are currently running.

        Returns:
            List[dict]: A List of dictionaries correlating a `Job` with its instance information.
            Each dictionary will be of the following format:
            {
                'job': <Job Instance>
                'id': <Instance's id>
                'start_time': <Datetime>
                'process': <Process instance>
            }
        """
        with self._executor_lock:
            return self._executor.get_instances()

    def add_listener(self, callback: Callable[[events.Event], None], mask: int = events.EVENT_ALL) -> None:
        """Adds a listener for an event.

        Args:
            callback (Callable): A function to call upon an event.
            mask (int, optional): Serves as a bitwise mask for the interested events. Defaults to
                events.EVENT_ALL.
        """
        with self._listeners_lock:
            self._listeners.append((callback, mask))

    def remove_listener(self, callback: Callable[[events.Event], None]) -> None:
        """Removes a listener.

        Args:
            callback (function): Used to find the listener to remove, if their callback functions
                match.
        """
        with self._listeners_lock:
            for i, (cb, _) in enumerate(self._listeners):
                if callback == cb:
                    del self._listeners[i]

    def _main_loop(self, event) -> None:
        """Processes the submitted `Job`s, unless in a non-running state."""
        max_wait = threading.TIMEOUT_MAX
        wait_seconds = max_wait
        while self.state != State.STOPPED:
            event.wait(wait_seconds)
            event.clear()
            if self.state == State.STOPPED:
                break
            wait_seconds = self._process_jobs()
            if wait_seconds == -1:
                wait_seconds = max_wait

    def _process_jobs(self) -> int:
        """Processes each `Job` that is currently due to run.

        After each `Job` execution its next scheduled run time is updated, or
        removed if it has none.

        Events produced by the `Job` executions also execute any requested listener
        callback functions.

        Returns:
            int: The number of seconds to wait until next wakeup.
        """
        if self.state == State.PAUSED:
            logger.debug('Scheduler paused; not processing jobs')
            return -1

        logger.info('Waking up to process jobs')
        now_dt = datetime.now(tz=self.tz)
        next_wakeup_time: Optional[datetime] = None
        events_to_dispatch = []

        with self._jobstore_lock:
            try:
                due_jobs = self._jobstore.get_due_jobs(now_dt)
                logger.debug(f'Due jobs: {due_jobs}')

                for job in due_jobs:
                    run_times = job._get_run_times(now_dt)
                    run_times = run_times[-1:] if job.coalesce else run_times
                    logger.debug(f'Job: {job}; run times: {run_times}')

                    for run_time in run_times:
                        try:
                            self._executor.submit_job(job, run_time)
                        except executor.MaxJobInstancesReachedException as e:
                            logger.error(f'Execution of job skipped: {e}')
                            events_to_dispatch.append(events.JobEvent(events.EVENT_JOB_MAX_INSTANCES, job.id))
                        except Exception as e:
                            logger.exception(e)
                        else:
                            events_to_dispatch.append(events.JobEvent(events.EVENT_JOB_SUBMITTED, job.id))

                    job_next_run_time = job.trigger.get_next_fire_time(run_times[-1], None)
                    logger.debug(f'Job next run time: {job_next_run_time}')
                    if job_next_run_time:
                        job.next_run_time = job_next_run_time
                        self._jobstore.update_job(job)
                    else:
                        self.remove_job(job.id)
            except Exception as e:
                logger.warning(f'Error getting due jobs from job store: {e}')
                next_wakeup_time = now_dt + timedelta(seconds=5)

            jobstore_next_run_time = self._jobstore.get_next_run_time()
            logger.debug(f'Jobstore next run time: {jobstore_next_run_time}')
            if jobstore_next_run_time is not None and next_wakeup_time is None:
                next_wakeup_time = jobstore_next_run_time

        for event in events_to_dispatch:
            self._dispatch_event(event)

        if self.state == State.PAUSED:
            wait_seconds = -1
            logger.debug('Scheduler is paused; waiting for resume')
        elif next_wakeup_time is None:
            wait_seconds = -1
            logger.debug('No jobs; waiting until a job is added')
        else:
            next_wakeup_time_sec = max((next_wakeup_time - now_dt).total_seconds(), 0)
            wait_seconds = min(int(next_wakeup_time_sec), int(threading.TIMEOUT_MAX))
            logger.debug(f'Next wakeup is due at {next_wakeup_time} (in {wait_seconds} seconds)')

        return wait_seconds

    def _dispatch_event(self, event: events.Event) -> None:
        """Given an event, execute any matching listener callbacks.

        Args:
            event (events.Event): The event that occurred.
        """
        with self._listeners_lock:
            listeners = tuple(self._listeners)

        for callback, mask in listeners:
            if event.code & mask:
                try:
                    callback(event)
                except Exception as e:
                    logger.error('Error notifying listener')
                    logger.exception(e)

    def _internal_listener(self, event: events.Event) -> None:
        """Internal `Scheduler` listener to watch for certain events.

        Args:
            event (events.Event): The event that occurred.
        """
        if isinstance(event, events.JobExecutionEvent):
            if event.code == events.EVENT_JOB_EXECUTED:
                with self._jobstore_lock:
                    logger.info(f'Next wakeup time: {self._jobstore.get_next_run_time()}')


class SchedulerAlreadyRunningException(Exception):
    """Raised when the `Scheduler` is already running."""
    def __init__(self) -> None:
        super().__init__('Scheduler is already running')


class SchedulerNotRunningException(Exception):
    """Raised when the `Scheduler` is not running."""
    def __init__(self) -> None:
        super().__init__('Scheduler is not running')


class JobAlreadyExistsException(Exception):
    """Raised when a `Job`'s unique ID already exists."""

    def __init__(self, job_id: str) -> None:
        """Initializes a JobAlreadyExistsException.

        Args:
            job_id (str): The job identifier that already exists.
        """
        super().__init__(f'Job already exists with ID {job_id}')
        self.job_id = job_id



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines `Trigger`s that can be fired to execute jobs on a given schedule.

Two default `Trigger`s are provided:
    `DateTrigger`: Which will fire at a given date and time.
    `CronTrigger`: Which acts like a typical cron job, accepting a cron expression.
"""

from datetime import datetime
from zoneinfo import ZoneInfo
from abc import ABC, abstractmethod
import logging
from typing import Optional
import zoneinfo
from croniter import croniter
from .util import get_class_logger


class Trigger(ABC):
    """Abstract Base Class that describes types of triggers for jobs.

    Attributes:
        logger (Logger): The class' logger.
        tz (zoneinfo): Timezone to use, set to UTC.
    """

    def __init__(self) -> None:
        self.logger: logging.Logger = get_class_logger(self)
        self.tz: zoneinfo = ZoneInfo('UTC')

    @abstractmethod
    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        """Returns the next scheduled time for the trigger to fire, if any.

        Args:
            previous (Optional[datetime]): Boundary datetime to look after.
                Defaults to None.
            latest (Optional[datetime]): Boundary datetime to look up until.
                Defaults to None.

        Returns:
            Optional[datetime]: The next scheduled time for the trigger. None if it isn't scheduled
                to be triggered.
        """
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class DateTrigger(Trigger):
    """A `Trigger` that is determined by a date and time.

    Attributes:
        run_time (datetime): When the trigger should be fired.
    """

    def __init__(self, run_time: datetime) -> None:
        """Initalizes a `DateTrigger`.

        Args:
            run_time (datetime): When the trigger should be fired.
        """
        super().__init__()
        self.run_time: datetime = run_time

    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        next_fire_time: Optional[datetime]
        next_fire_time = self.run_time if previous is None or self.run_time > previous else None
        return next_fire_time

    def __getstate__(self) -> dict[str, Optional[datetime]]:
        """Returns the current state, used for pickling.

        Returns:
            dict: Dictionary mapping of `DateTrigger` attribute to value.
        """
        return {
            'run_time': self.run_time
        }

    def __setstate__(self, state: dict[str, datetime]) -> None:
        """Explicitly sets all the attributes, used for unpickling.

        Args:
            state (Dict): Mapping of `DateTrigger` attributes to their respective values.
        """
        self.run_time = state['run_time']

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} (run_time="{self.run_time}")>'

    def __str__(self) -> str:
        return f'date[{self.run_time}]'


class CronTrigger(Trigger):
    """A `Trigger` setup as a cron job.

    Attributes:
        cron_str (str): A cron expression string, five time units can be specified as follows:
            <minute> <hour> <day of month> <month> <day of week> 
    """

    def __init__(self, cron_str: str) -> None:
        """Initializes a CronTrigger.

        Args:
            cron_str (str): A cron expression string, five time units can be specified as follows:
                <minute> <hour> <day of month> <month> <day of week> 
        """
        super().__init__()
        self.cron_str: str = cron_str

    def get_next_fire_time(self, previous: Optional[datetime] = None, latest: Optional[datetime] = None) -> Optional[datetime]:
        if previous is None:
            previous = datetime.now(tz=self.tz)

        itr = croniter(self.cron_str, previous)

        next_fire_time = datetime.fromtimestamp(itr.get_next(), tz=self.tz)

        if latest is not None and next_fire_time > latest:
            return None

        return next_fire_time

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} (cron_str="{self.cron_str}")>'

    def __str__(self) -> str:
        return f'cron[{self.cron_str}]'



# Copyright 2020 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup, find_packages

my_dir = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(my_dir, 'README.md')) as f:
    README = f.read()

with open(f'{my_dir}/requirements.txt') as file:
    REQUIRES = file.readlines()

with open(f'{my_dir}/requirements-dev.txt') as file:
    REQUIRES_DEV = file.readlines()

setup(
    name='yapsched',
    version='0.3.0',
    description='Yet Another Python Scheduler',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
    ],
    author='Software Factory Labs',
    author_email='eric@softwarefactorylabs.com',
    packages=find_packages(include=['yapsched', 'yapsched.*']),
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIRES,
    extras_require={
        'dev': REQUIRES_DEV,
    },
    python_requires='>= 3.9, < 4',
)
